module ccm_barrier_mod
    contains
    subroutine ccm_barrier(wait,io_flush,the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        real,optional :: wait
        logical,optional :: io_flush
        integer,optional :: the_err
        integer :: c,r,m,it
        real(b8):: t1,t2,dt,new_t,tx1,tx2
        if(present(the_err))the_err=0
        call shmem_barrier_all;if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
        if(present(io_flush))then
            if(io_flush)then
!               do i=1,size(ccm_iounits)
!                   call flush(ccm_iounits(i))
!               enddo
                call shmem_barrier_all;if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
            endif
        endif
        if(present(wait))then
            dt=wait
        else
            return
        endif
        tx1=ccm_time()+dt
        if(dt .gt. 0 .and. dt .lt. 1.0_b8)then
            call system_clock(count=c,count_rate=r,count_max=m)
            t1=real(c,b8)/real(r,b8)
            t2=t1+dt
            new_t=t1
            do while(t2 .gt. new_t)
                call system_clock(count=c)
                new_t=real(c,b8)/real(r,b8)
                if(new_t .lt. t1)then
                    new_t=real(m,b8)-new_t
                    t1=new_t
                    t2=real(m,b8)-t2
                endif
            enddo
        else
            if(dt .gt. 0.0_b8)call sleep(nint(dt))
        endif
        tx2=ccm_time()
        do while(tx2 .lt. tx1)
                tx2=ccm_time()
        enddo
        call shmem_barrier_all;if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
    end subroutine
    
    subroutine ccm_probe(flag,status)
        use ccm_numz
        use ccm_error_mod
        implicit none
        logical :: flag
        integer :: status(2)
        integer :: i
        sync_block(myid)=-1
        do i=0,numnodes-1
        	if(sync_block(i) .ne. -1)then
        		flag=.true.
        		status(1)=i
        		status(2)=0
        		return
        	endif
        enddo
		flag=.false.
		status(1)=-1
		status(2)=-1
 end subroutine
 
subroutine ccm_probe_list_init(the_list,the_root)
		use ccm_numz
		type(node_list) :: the_list(:)
		integer:: the_root
		do i=-1,numnodes-1
			the_list(i)%from=(i-1)
			the_list(i)%to=(i+1)
			the_list(i)%done=.false.
		enddo
		the_list(-1)%from=numnodes-1
		the_list(-1)%done=.true.
		the_list(numnodes-1)%to=-1
	! cut root out of the linked list
		the_list(the_root)%done=.true.
		j=the_list(the_root)%to
		i=the_list(the_root)%from
		the_list(i)%to=j
		the_list(j)%from=i
end subroutine ccm_probe_list_init
subroutine ccm_probe_list(the_list,flag,status)
		use ccm_numz
		type(node_list) :: the_list(:)
		logical :: flag
		integer :: status(2)
		integer :: i,j,next,last
		i=-1
		i=the_list(i)%to
		if(i .eq. -1)then
			flag=.false.
			status(1)=-1
			status(2)=0
			return
		endif
		do
			if(sync_block(i) .ne. -1)then
			! message waiting
				flag=.true.
				status(1)=i
				status(2)=0
				!cut me out of the loop
				the_list(i)%done=.true.
				next=the_list(i)%to
				last=the_list(i)%from
				the_list(next)%from=last
				the_list(last)%to=next
				return
			endif
			next=the_list(i)%to
			if(i .eq. next)then
				flag=.false.
				status(1)=-1
				status(2)=0
				return
			endif
			i=next
		enddo
end subroutine ccm_probe_list
    subroutine ccm_checkin(wait,name,the_err)
	use ccm_numz
	use ccm_error_mod
	implicit none
	real,optional,intent(in) :: wait
	character(len=*),intent(in),optional :: name
	integer, optional :: the_err
	real(b8) :: wtime,st,tn,et
	integer :: ierr,igot,i,len,next
	logical :: flag,not_done
	character(len=32)  :: lname
	integer(selected_int_kind(18)) :: mychar(4),m1,p1,m2,z0
	integer :: exit_case
	save
	if(present(the_err))the_err=0
	if(present(wait))then
		if(wait .le. 0)return
		wtime=wait
	else
		wtime=10.0
	endif
	if(present(name))then
		if(len_trim(name) .le. 32)then
			write(lname,"(a)")trim(name)
		else
			write(lname,"(a)")name(1:32)
		endif
	else
		write(lname,"(a)")"ccm_undefined"
	endif
	m1=-1
	p1=1
	z0=0
	m2=-2
	lname=trim(lname)
	len=len_trim(lname)
	if(len .lt. 32)lname(len+1:32)=' '
	st=ccm_time()
	tn=st
	et=st+wtime
	mychar=0
	mychar=transfer(lname,mychar)
	if(myid .eq. 0)then
		myflag=0
		next=1
		call shmem_put64(mycray(1:4,0),mychar,4,next)
		call shmem_quiet
		call shmem_put64(myflag(0),p1,1,next)
		call shmem_quiet
        not_done=.true.
        !write(*,*)st,et
        do while (not_done)
        	if(not_done)then
        		if(myflag(0) .eq. p1)then
        			exit_case=1
        			not_done=.false.
        		endif
        	endif
        	if(not_done)then
				if(myflag(0) .eq. m1)then
					exit_case=2
					not_done=.false.
					!write(*,*)"exit_case=2"
				endif
        	endif
        	if(not_done)then
 				tn=ccm_time()
 				!write(*,*)"what ",tn,et
 				if(tn .ge. et)then
 				    !write(*,*)myid," what ",tn,et
        			exit_case=3
        			not_done=.false.
        		endif
        	endif
        enddo
       !write(*,*)myid,"exit case ",exit_case
        if(exit_case .eq. 1)then
			myflag(myid)=0
			if(present(the_err))the_err=0
			do i=1,numnodes-1
				call shmem_put64(myflag(i),p1,1,i)
				call shmem_quiet
			enddo
			!write(*,*)myid,"calling shmem_barrier_all"
			call shmem_barrier_all
			myflag=0
			mycray(1:4,0)=0
			mychar=0
			call shmem_barrier_all
			return
		endif
        if(exit_case .eq. 2)then
			myflag(myid)=0
			if(present(the_err))the_err=0
			do i=1,numnodes-1
				call shmem_put64(myflag(i),m2,1,i)
				call shmem_quiet
			enddo
			if(present(the_err))the_err=-1
			write(err_str1,*)"deadlock detected at ",tn," process ",myid
			write(err_str2,*)"waiting for label ",trim(lname)
			call ccm_warning("ccm_checkin")
			return
        endif
        if(exit_case .eq. 3)then
			do i=1,numnodes-1
				call shmem_put64(myflag(i),m1,1,i)
				call shmem_quiet
			enddo
			if(present(the_err))the_err=-1
			write(err_str1,*)"process ",myid," timed out at ",tn
			write(err_str2,*)"waiting for label ",trim(lname)
			call ccm_warning("ccm_checkin")
        endif
        myflag(0)=0
	endif
	if(myid .ne. 0)then
		st=ccm_time()
        tn=st
        et=st+wtime
        not_done=.true.
        do while (not_done)
        	if(not_done)then
				if(myflag(0) .eq. p1)then
					not_done=.false.
					myflag(0)=0
					myflag(myid)=0
					if(sum(abs(mychar(1:4)-mycray(1:4,0))) .eq. 0)then
					! i am ok, send the message on
						next=myid+1
						if(next.eq. numnodes)next=0
						call shmem_put64(mycray(1:4,0),mychar,4,next)
						call shmem_quiet
						call shmem_put64(myflag(0),p1,1,next)
						call shmem_quiet
						exit_case=1
					else
					! send error to zero
						myflag(0)=0
						call shmem_put64(myflag(myid),m1,1,0)
						call shmem_quiet
						exit_case=2
						next=myid+1
					! send error to next node
						if(next.ne.  numnodes)then
							call shmem_put64(myflag(0),m1,1,next)
							call shmem_quiet
						endif
					endif
				endif
        	endif
        	if(not_done)then
 				tn=ccm_time()
 				if(tn .ge. et)then
        			exit_case=3
        			not_done=.false.
        		endif
        	endif
        	if(not_done)then
 				if(myflag(0) .eq. m1)then
					myflag(0)=0
					next=myid+1
					not_done=.false.
					exit_case=2
					if(next.ne.  numnodes)then
						call shmem_put64(myflag(0),m1,1,next)
						call shmem_quiet
					endif
				endif
        	endif
        enddo
        !write(*,*)myid,"exit case ",exit_case
			if(exit_case .eq. 1)then
			! i am ok wait for confirmation on second phase
			! we either get p1 = ok or m1 = error from zero
				CALL SHMEM_INT8_WAIT(myflag(myid),z0)
				if(myflag(myid) .eq. p1)then
					if(present(the_err))the_err=0
					!write(*,*)myid,"calling shmem_barrier_all"
					call shmem_barrier_all
					myflag=0
					mycray(1:4,0)=0
					mychar=0
					call shmem_barrier_all
					return
				endif
				if(myflag(myid) .eq. m1)then
					myflag(myid)=0
					if(present(the_err))the_err=-1
					!write(*,*)myid," what ",tn,et
					write(err_str1,*)"process ",myid," timed out at ",tn
					write(err_str2,*)"waiting for label ",trim(lname)
					call ccm_warning("ccm_checkin")
				endif
				if(myflag(myid) .eq. m2)then
					myflag(myid)=0
					if(present(the_err))the_err=-1
					write(err_str1,*)"deadlock detected at ",tn," process ",myid
					write(err_str2,*)"waiting for label ",trim(lname)
					call ccm_warning("ccm_checkin")
				endif
			endif
			if(exit_case .eq. 2)then
				myflag(myid)=0
				if(present(the_err))the_err=-1
				write(err_str1,*)"deadlock detected at ",tn," process ",myid
				write(err_str2,*)"waiting for label ",trim(lname)
				call ccm_warning("ccm_checkin")
			endif
			if(exit_case .eq. 3)then
				myflag(myid)=0
				if(present(the_err))the_err=-1
				!write(*,*)myid," what ",tn,et
				write(err_str1,*)"process ",myid," timed out at ",tn
				write(err_str2,*)"waiting for label ",trim(lname)
				call ccm_warning("ccm_checkin")
			endif
        endif
end subroutine


end module
