module ccm_init_mod
 logical,private,save :: ccm_i_did_it
 logical,private,save :: ccm_open =.false.
 integer,private,parameter :: ccm_ntypes=9
 integer,private,save :: ccm_sizes(ccm_ntypes),ccm_kinds(ccm_ntypes)
 character(len=32),private,save :: ccm_names(ccm_ntypes)
private bytes
    contains 
    function bytes(digits)
        implicit none
        integer,intent(in) :: digits
        integer :: bytes,temp
        temp=ceiling(real(digits)/8.0)
        if(temp .gt. 8)then
            bytes=16
            return
        endif
        if(temp .gt. 4)then
            bytes=8
            return
        endif
        if(temp .gt. 2)then
            bytes=4
            return
        endif
        if(temp .gt. 1)then
            bytes=2
            return
        endif
        bytes=0
    end function
    
    subroutine ccm_unique(name,base)
        use ccm_numz
        implicit none
        character (len=*) ::  name
        character (len=*), optional ::  base
        character (len=80) :: temp
        if(present(base))then
            write(temp,"(a,i5.5)")trim(base),myid
        else
            write(temp,"(a,i5.5)")trim("out_"),myid
        endif
        name=trim(temp)
    end subroutine ccm_unique
    
    subroutine ccm_info(my_id,num_procs,n_types,sizes,kinds,names,the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        integer,optional,intent(out) :: my_id,num_procs,n_types
        integer,optional,intent(out) :: sizes(:),kinds(:)
        character(len=20),optional,intent(out) :: names(:)
        integer,optional,intent(out) :: the_err
        if(present(the_err))the_err=0

        if(present(my_id))my_id=myid
        if(present(num_procs))num_procs=numnodes
        if(present(n_types))n_types=ccm_ntypes
        
        if(present(sizes))then
        	if(size(sizes) .lt. ccm_ntypes)then
        	    write(err_str1,*)"sizes array too small needs to be : ",ccm_ntypes
                call ccm_warning("ccm_info")
                if(present(the_err))the_err=1
                return
        	else
        	    sizes(1:ccm_ntypes)=ccm_sizes(1:ccm_ntypes)
        	endif
        endif
        if(present(kinds))then
        	if(size(kinds) .lt. ccm_ntypes)then
        	    write(err_str1,*)"kinds array too small needs to be : ",ccm_ntypes
                call ccm_warning("ccm_info")
                if(present(the_err))the_err=1
                return
        	else
        	    kinds(1:ccm_ntypes)=ccm_kinds(1:ccm_ntypes)
        	endif
        endif
        if(present(names))then
        	if(size(names) .lt. ccm_ntypes)then
        	    write(err_str1,*)"names array too small needs to be : ",ccm_ntypes
                call ccm_warning("ccm_info")
                if(present(the_err))the_err=1
                return
        	else
        	    names(1:ccm_ntypes)=ccm_names(1:ccm_ntypes)
        	endif
        endif
    end subroutine ccm_info
        
    

    subroutine ccm_init(my_id,num_procs,info,the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        integer :: my_pe, num_pes
        real :: normal_real
        complex :: normal_complex
        double precision :: normal_dp
        integer :: normal_integer
        integer(def_int) :: my_integer
        logical :: normal_logical
        character :: normal_char
        integer :: lr,lc,ld,li,la,ll,di
        
        integer, optional,intent (out) :: the_err
        integer, optional,intent (out) :: my_id
        integer, optional,intent (out) :: num_procs
        logical, optional,intent (in) :: info
        logical :: flag
        logical, save :: been_called=.false.
        integer :: real_extent,double_extent,comp_extent,dpcp_extent,int_extent
        integer :: lshort,llong,lcomp,ldpcp,lint
        real(b4):: short_x
        real(b8):: long_x
        complex(c4):: short_c
        complex(c8):: long_c
        if(present(the_err))the_err=MPI_ZERO
        call ccm_clear_warning()
        if(been_called)then
            write(err_str1,*)"ccm_init has already been called"
            call ccm_fatal("ccm_init")
            return
        endif
        call start_pes(0)
        myid = my_pe()
        numnodes=num_pes()
        sync_block=-1
        mycray=0
        myflag=0
        been_called =.true.
        ccm_auto_print=.true.
        do_tests=0
        call date_and_time(values=ccm_start_time)
        ccm_i_did_it=.true.
        ccm_open=.true.
        allocate(got_list(-1:numnodes-1))

! data types are defined in ccm_numz
!        mydouble
!        myint
!        mylogical
!        mycharacter
!        myreal
!        mycomp
!        mydpcomp
        
        if(mpi_err .ne. 0)call ccm_mpi_error("ccm_init")
! check the match between fortran sizes and mpi sizes
!        call MPI_TYPE_EXTENT(myreal,real_extent,mpi_err)
        short_x=0.0
        lshort=bytes(digits(short_x))
!
!        call MPI_TYPE_EXTENT(mydouble,double_extent,mpi_err)
        long_x=0.0
        llong=bytes(digits(long_x))
!
!        call MPI_TYPE_EXTENT(mycomp,comp_extent,mpi_err)
        short_c=0.0
        lcomp=bytes(digits(real(short_c,c4)))*2
!       
!        call MPI_TYPE_EXTENT(mydpcomp,dpcp_extent,mpi_err)
        long_c=0.0
        ldpcp=bytes(digits(real(long_c,c8)))*2
!       
!        call MPI_TYPE_EXTENT(myint,int_extent,mpi_err)
        my_integer=0
        lint=bytes(digits(my_integer))
        if(mpi_err .ne. 0)call ccm_mpi_error("ccm_init")
!       
        normal_real=0
        normal_complex=0
        normal_dp=0
        normal_integer=0
        normal_logical=.false.
        normal_char=char(32)
        lr=bytes(digits(normal_real))
        lc=bytes(digits(real(normal_complex)))
        ld=bytes(digits(normal_dp))
        li=bytes(digits(normal_integer))
        ll=1  !portable?
        la=1  !portable?
        if(present(my_id))my_id=myid
        if(present(num_procs))num_procs=numnodes
        call ccm_unique(ccm_name,"ccm_out_")
		ccm_sizes(1 )=bytes(digits(short_x))
		ccm_kinds(1 ) =kind(short_x)
		ccm_names(1 ) ="real short"
		
		ccm_sizes(2 )=bytes(digits(long_x))
		ccm_kinds(2 ) =kind(long_x)
		ccm_names(2 ) ="real long"
		
		ccm_sizes(3 )=bytes(digits(real(short_c,c4)))*2
		ccm_kinds(3 ) =kind(short_c)
		ccm_names(3 ) ="complex short"
		
		ccm_sizes(4 )=bytes(digits(real(long_c,c8)))*2
		ccm_kinds(4 ) =kind(long_c)
		ccm_names(4 ) ="complex long"
		
		ccm_sizes(5 )=bytes(digits(my_integer))
		ccm_kinds(5 ) =kind(my_integer)
		ccm_names(5 ) ="integer default"
		
		ccm_sizes(6 )=ll
		ccm_kinds(6 ) =kind(normal_logical)
		ccm_names(6 ) ="logical default"
		
		ccm_sizes(7 )=la
		ccm_kinds(7 ) =kind(normal_char)
		ccm_names(7 ) ="character default"
		
		if(ccm_ntypes .eq. 9)then
			if(b16 .gt. 0)then
			ccm_sizes(8 )=ccm_sizes(2 )*2
			ccm_kinds(8 )=b16
			ccm_names(8 )="quad real"
		
			ccm_sizes(9 )=ccm_sizes(2 )*4
			ccm_kinds(9 )=c16
			ccm_names(9 )="quad complex"
			else
			ccm_sizes(8 )=-1
			ccm_kinds(8 )=-1
			ccm_names(8 )="quad real undefined"
		
			ccm_sizes(9 )=-1
			ccm_kinds(9 )=-1
			ccm_names(9 )="quad complex undefined"
			
			endif
                	
        endif
        if(present(info))then
            if(info .and. myid .eq. 0)then
                write(*,*)"Collective Communications Module"
                write(*,*)"  SHMEM reference implementation"
                write(*,fmt="(""real size =      "",i4)")lr
                write(*,fmt="(""complex size =   "",i4)")lc
                write(*,fmt="(""double size =    "",i4)")ld
                write(*,fmt="(""integer size =   "",i4)")li
                write(*,fmt="(""character size = "",i4)")la
                write(*,fmt="(""logical size =   "",i4)")ll
                write(*,fmt="(""maximum number of processors =   "",i4)")maxprocs
            endif
        endif
        call shmem_barrier_all
        if(numnodes .gt. maxprocs)then
        	if(myid .eq. 0)then
        		write(*,fmt="(""maximum number of processors =   "",i4)")maxprocs
        		write(*,*)"please rebuild ccm_com with a larger value of maxprocs"
        	endif
        	stop
        endif
    end subroutine
    subroutine ccm_close(the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        integer, optional,intent (out) :: the_err
        if(.not. ccm_open)then
        	write(err_str1,*)"ccm_init is not open"
            call ccm_warning("ccm_close")
            if(present(the_err))the_err=-1
         	return
        endif
        ccm_open=.false.
    end subroutine ccm_close
end module
